/*********************************************************************************
 *
 * Razor! Engine - A modular C++ presentation engine
 *
 * $Id: CollisionDetector.cpp,v 1.1 2003/02/15 11:35:32 teacy Exp $
 *
 * Copyright (c) 2000-2003 Tilo Christ. All Rights Reserved.
 * Copyright (c) 1994 Tom Moertel (moertel@acm.org)
 *
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a 
 * copy of this software and associated documentation files (the  "Software"), 
 * to deal in the Software without restriction, including without limitation 
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the Software 
 * is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all 
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
 * INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
 * PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE 
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, 
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 **********************************************************************************/


// This code was inspired by the artice at http://gpp.netfirms.com/cgi-bin/resource.cgi?c_collision1


// Sector-based collision detection routines 
//
// Tom Moertel 21-Jun-94
//
// Results for a 25 MHz 68040 Macintosh (not
// exactly a screamer) and an 80 MHz PPC 601
// Power Macintosh 8100 (this one screams):
//
//                          tests/s
//   object count        -68K-  -PPC-
//
//        0               611   7640
//       50               340   4020
//      100               189   2060
//      200                81    788
//
// where a "test" is defined to be a complete
// check of all objects, determining for each
// object whether it is involved in a collision
// (and if it is, with what other object).
//
// NOTES
//
// For this job I made all objects 30x30, but
// the code will work for arbitrarily-sized
// objects, with the restriction that objects
// are smaller than half of kSectorSize.
//
// This code is far from optimized.  I didn't
// even bother to run it through a profiler.
// With a little work, it could probably be
// twice as fast.
//
// LEGAL STUFF
//
// Feel free to use this code in your own
// projects, but please give me credit.
//
// Copyright 1994 by Tom Moertel
// moertel@acm.org
//

// define compilation parameters

#if defined(__MWERKS__)
#define BRAIN_DEAD_INLINING     // define this to declare "hot"
#endif                          // functions as macros instead
                                // of C++ inline functions

// define test parameters

enum
{
    kMaxObjects     = 200,      // more than you're likely to need
    kRectSize       = 30,       // each object is 30 x 30 pixels
};


// sprite object

struct obj
{
    Coord   x, y;           // coords
    obj*    sector_link;    // link in sector list
    obj*    obj_link;       // link in obj list
} ;


// module-scope globals

static obj      gObjects[kMaxObjects];
static Boolean  gCollisionArray[kMaxObjects];

// forward declatations

static void _DetermineCollisions();


//==================================================================
// sector code
//==================================================================

#define CEILING_DIV(x, y) ( ((x)+(y)-1) / (y) )

// define constants
//
// Note that to work properly, kSectorSize must be greater
// than twice the length of the largest side of any
// object's bounding box.  E.g., if your objects are
// 30x30, then the sector size should be > 60 -- 64 would
// be an excellent choice.

enum {
    kSectorSize     = 64,   // length of a sector's side in pixels
    kLog2SectorSize =  6,   // log2(kSectorSize): for shifting
    
    kScreenWidth    = 640,
    kScreenHeight   = 480,
    
    kNumXSectors    = CEILING_DIV(kScreenWidth, kSectorSize) + 1,
    kNumYSectors    = CEILING_DIV(kScreenHeight, kSectorSize) + 1,
    kNumSectors     = kNumXSectors * kNumYSectors
} ;

// define a module-scope array of linked list heads,
// one for each sector

static obj* gSectorArray[kNumXSectors][kNumYSectors];


// call this routine to place all objects into the
// appropriate sectors
//
// (assumes all objects are kept in a linked list and
// GetMyFirstObject() returns the head of this list)

extern obj* GetMyFirstObject();

static void UpdateSectors(register scalar xoff, register scalar yoff)
{
    // reset the sectors' linked lists
    
    obj** theArray = (obj**) gSectorArray; // for 1-D access
    for (scalar i = 0; i < kNumSectors; i++)
        *theArray++ = NULL;
    
    // put each object in its sector's linked list.
    
    for (obj* o = GetMyFirstObject(); o != NULL; o = o->obj_link)
    {
        // get the list head for the sector in which o resides
    
        register obj** thisSectorListHead =
            &gSectorArray [ (o->x + xoff) >> kLog2SectorSize ]
                          [ (o->y + yoff) >> kLog2SectorSize ];
        
        // add o to this sector's linked list
        
        o->sector_link = *thisSectorListHead;
        *thisSectorListHead = o;
    }
}



static void _RandomizeObjects(scalar numObjects)
{
    obj* o = gObjects;

    for (scalar i = 0; i < numObjects; i++, o++)
    {
        o->x        = _RandScalar(kScreenWidth-1);
        o->y        = _RandScalar(kScreenHeight-1);
        o->obj_link = o + 1;
    }
    
    (--o)->obj_link = NULL;
}


//==================================================================
// intersection code
//==================================================================

obj* GetMyFirstObject() { return &gObjects[0]; }

// local helpers

static void _ClearCollisionArray();
static void _UpdateCollisionArray();

// determine all collisions

static void _DetermineCollisions()
{
    _ClearCollisionArray(); // erase the slate; no collisions yet

    scalar shift = kSectorSize / 2;
    
    // We need to try four differnt "shifts" of the
    // sector grid to detect all collisions.  Proof of
    // why this is so is left as an excercise for the
    // reader.  (Hint: consider an analogous 1-D case.)
    
    UpdateSectors(    0,     0),    _UpdateCollisionArray();
    UpdateSectors(    0, shift),    _UpdateCollisionArray();
    UpdateSectors(shift,     0),    _UpdateCollisionArray();
    UpdateSectors(shift, shift),    _UpdateCollisionArray();
}

// "hot" functions that are used in inner loops

#ifdef BRAIN_DEAD_INLINING

#define _Abs(a) ((a) < 0 ? -(a) : (a))

#define _IntersectQ(o1, o2)                     \
            (_Abs(o1->x - o2->x) < kRectSize && \
             _Abs(o1->y - o2->y) < kRectSize)
#else

inline scalar _Abs(scalar a)
{
    return a < 0 ? -a : a;
}

inline scalar _IntersectQ(obj* o1, obj* o2)
{
    return _Abs(o1->x - o2->x) < kRectSize &&
           _Abs(o1->y - o2->y) < kRectSize;
}

#endif // BRAIN_DEAD_INLINING


static void _ClearCollisionArray()
{
    memset(gCollisionArray, 0, sizeof(gCollisionArray));
}

static void _CalcCollisionsInSector(obj* objList);

static void _UpdateCollisionArray()
{
    for (scalar x = 0; x < kNumXSectors; x++)
        for (scalar y = 0; y < kNumYSectors; y++)
            _CalcCollisionsInSector(gSectorArray[x][y]);
}


// We've got the head of the linked list for a 
// sector.  Let's see if there are any objects
// in it that are involved in collisions.
//
// Use the plain, old O(n^2) technique to compute
// the collisions in this sector.  If the grid size
// was appropriately chosen, n should be very small;
// in many cases it will be 0 or 1, obviating
// collision tests altogether.

static void _CalcCollisionsInSector(obj* objList)
{
    if (objList == NULL || objList->sector_link == NULL)
    {
        return;
    }

    for (obj* o0 = objList; o0->sector_link; o0 = o0->sector_link)
    {
        for (obj* ox = o0->sector_link; ox; ox = ox->sector_link)
        {
            if (_IntersectQ(o0, ox))
            {
                gCollisionArray[ o0 - gObjects ] =
                gCollisionArray[ ox - gObjects ] = 1;
    		}
                // Note that at this point we know object o0
                // collided with object ox, so we could use that
                // information to, say, determine what kind of
                // explosion is appropriate.  Here, however, I
                // just toss the information away.
		}
	}
}
