/*********************************************************************************
 *
 * Razor! Engine - A modular C++ presentation engine
 *
 * $Id: Presentation.cpp,v 1.4 2003/05/31 20:05:28 teacy Exp $
 *
 * Copyright (c) 2000-2003 Tilo Christ. All Rights Reserved.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a 
 * copy of this software and associated documentation files (the  "Software"), 
 * to deal in the Software without restriction, including without limitation 
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the Software 
 * is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all 
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
 * INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
 * PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE 
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, 
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 **********************************************************************************/

#include "./HostControl.h"
#include <PalmOS.h>

#include "Utilities.h"
#include "Device.h"
#include "app/Presentation.h"
#include "gfx/DisplayFactory.h"
#include "input/HardKeyManager.h"
#include "sound/SoundEngine.h"

#include "MyActionEngine.h"


using namespace prvrazor;

Presentation::Presentation() :
	display(DisplayFactory::createDisplay()), actionEngine(*(new MyActionEngine()))
{
	HardKeyManager::init();
	SoundEngine::init();
		
	state    = presReady;
	imgState = imgCorrupt;

    currentScreen = NULL;
		
	// Initialize the ActionEngine
	actionEngine.init();

	// Restore saved state
#ifndef MODE_PROFILING
	UInt16 prefsSize = 0;
	Int16 prefResult = PrefGetAppPreferences(Device::creatorID, 0, NULL, &prefsSize, true);
		
	if (prefResult != noPreferenceFound)
	{
		UInt16 prefsVersion;
		UInt16 prefsSize = sizeof(UInt16);
		PrefGetAppPreferences(Device::creatorID, 0, &prefsVersion, &prefsSize, true);

		if (prefsVersion == actionEngine.getStateDescriptorsVersion())
		{
			// Restore preferences
			UInt16 index = 0;
			while(true)
			{
				StateDescriptor *stateDescriptor = actionEngine.getStateDescriptor(index);
				if (stateDescriptor == NULL)
				{
					break;
				}
				
				UInt16 prefsSize = stateDescriptor->size;
				PrefGetAppPreferences(Device::creatorID, index + 1, stateDescriptor->state, &prefsSize, true);
				
				index++;
			}
		}
	} 
#endif
}

	
Presentation::~Presentation()
{
#ifndef MODE_PROFILING
	// Save state
	const UInt16 prefsSize = sizeof(UInt16);
	const UInt16 prefsVersion = actionEngine.getStateDescriptorsVersion();
	PrefSetAppPreferences(Device::creatorID, 0, Device::appVersionNum, &prefsVersion, prefsSize, true);

	// Save preferences from ActionEngine
	UInt16 index = 0;
	while(true)
	{
		StateDescriptor *stateDescriptor = actionEngine.getStateDescriptor(index);
		if (stateDescriptor == NULL)
		{
			break;
		}
		
		PrefSetAppPreferences(Device::creatorID, index + 1, Device::appVersionNum, stateDescriptor->state, stateDescriptor->size, true);
		
		index++;
	}
#endif

	delete currentScreen;
	delete &actionEngine; 
	display.destroy();
	delete &display;
	HardKeyManager::destroy();
	SoundEngine::destroy();
}


void Presentation::begin()
{
	HostTraceOutput("Presentation::begin()");

	HardKeyManager::captureHardKeys();

	// Initialize first screen
	currentScreen = actionEngine.getCurrentScreen();
	DisplayProperties displayProperties;
	currentScreen->getDisplayProperties(displayProperties);
	display.initDisplay(displayProperties);

	ViewportProperties viewport;
	currentScreen->getViewportProperties(display.getDisplayCanvas(), viewport);	
	display.initViewport(viewport);

	display.captureDisplay();

	currentScreen->init(display.getDisplayCanvas(), display.getViewportCanvas());
	advanceTimeInterval = currentScreen->getAdvanceTimeInterval();
	pauseLengthBeforeResuming = currentScreen->getPauseLengthBeforeResuming();

	state = presResuming;
	pausedTime = 0;

	// Synchronize the periods to the timer.
	nextPeriodTime = TimGetTicks() + max(pauseLengthBeforeResuming, advanceTimeInterval);

	this->redraw();
}


Boolean Presentation::hasBegun() const
{
	return (state != presReady);
}	
	

void Presentation::nextPeriod()
{
//	HostTraceOutput("Presentation::nextPeriod()");

	state = presRunning;

	nextPeriodTime = TimGetTicks() + advanceTimeInterval;		
		
	// Prevent auto-off
	if (actionEngine.getPreventAutoOff())
	{
		EvtResetAutoOffTimer();
	}
	
	// Show display buffer on screen
	this->show();				

	// Make sure SoundEngine gets invoked on a regular basis
	SoundEngine::timeTick();

	// Move objects, etc...
	if (currentScreen->nextPeriod())
	{
		// The current screen has requested to be exited. Switch to next one...
		switchScreen();
	}
	
	// Render into display-buffer
	this->draw();				

	// Make sure SoundEngine gets invoked on a regular basis
	SoundEngine::timeTick();
}


void Presentation::redraw()
{
	HostTraceOutput("Presentation::redraw()");

	imgState = imgCorrupt;
	this->show();
}

	
Boolean Presentation::isPaused() const
{
	return (state == presPaused);
}	


void Presentation::pause()
{
	HostTraceOutput("Presentation::pause()");

	if (state == presPaused)
	{
		return;
	}
	
	state = presPaused;
	pausedTime = TimGetTicks();

	display.releaseDisplay();
	HardKeyManager::releaseHardKeys();	
}


void Presentation::resume()
{
	HostTraceOutput("Presentation::resume()");

	HardKeyManager::captureHardKeys();
	display.captureDisplay();
	
	this->redraw();

	if (state == presPaused)
	{
		nextPeriodTime += (TimGetTicks() - pausedTime);

		state = presResuming;
	}
	else
	{
		nextPeriodTime = TimGetTicks() + advanceTimeInterval;
	}					
}
	

Int32 Presentation::getTimeUntilNextPeriod() const
{
	if (state == presPaused)
	{
		return evtWaitForever;
	}
	else
	{	
		Int32 timeRemaining = (Int32)(nextPeriodTime - TimGetTicks());
		if (timeRemaining < 0)
		{
			timeRemaining = 0;
		}

		return timeRemaining;	
	}
}


#ifndef MODE_PROFILING
void Presentation::receiveEvent(const EventType* evtPtr)
{
	currentScreen->receiveEvent(evtPtr);
}
#else
void Presentation::receiveEvent(const EventType* /* evtPtr */)
{
}
#endif
	

void Presentation::draw()
{		
//	HostTraceOutput("Presentation::draw()");
	
	ErrNonFatalDisplayIf(imgState == imgViewportReady, "Presentation.draw: Duplicate draw");

	RectangleType bounds;
	bounds.topLeft.x = bounds.topLeft.y = 0;
	bounds.extent.x = bounds.extent.y = 0;

	if (imgState == imgCorrupt)
	{
		display.beginDrawViewport();
		currentScreen->drawBackground(display.getViewportCanvas());
		currentScreen->drawAction(display.getViewportCanvas(), &bounds);
		display.endDrawViewport(NULL);
		currentScreen->drawFixedOverlays(display.getDisplayCanvas());
	}
	else
	{
		display.beginDrawViewport();
		currentScreen->drawAction(display.getViewportCanvas(), &bounds);
	
		display.endDrawViewport(&bounds);
	}		

	imgState = imgViewportReady;
}
	
	
void Presentation::show()
{
//	HostTraceOutput("Presentation::show()");
	
	if (imgState != imgViewportReady)
		this->draw();
	
	display.showViewport();

	currentScreen->drawDynamicOverlays(display.getDisplayCanvas());

	imgState = imgViewportShowing;
}



void Presentation::switchScreen()
{
	imgState = imgCorrupt;
	delete currentScreen;
	currentScreen = actionEngine.getNextScreen();	

	DisplayProperties displayProperties;
	currentScreen->getDisplayProperties(displayProperties);
	display.initDisplay(displayProperties);

	ViewportProperties viewport;
	currentScreen->getViewportProperties(display.getDisplayCanvas(), viewport);	
	display.initViewport(viewport);

	currentScreen->init(display.getDisplayCanvas(), display.getViewportCanvas());
	advanceTimeInterval = currentScreen->getAdvanceTimeInterval();
	pauseLengthBeforeResuming = currentScreen->getPauseLengthBeforeResuming();

	currentScreen->nextPeriod();
}



bool Presentation::enableMenu()
{
    if (currentScreen == NULL)
    {
        return false;
    }
    else
    {
        return currentScreen->enableMenu();
    }
}
