/*********************************************************************************
 *
 * Razor! Engine - A modular C++ presentation engine
 *
 * $Id: Direct4bppBlitter.cpp,v 1.3 2003/03/17 22:29:20 teacy Exp $
 *
 * Copyright (c) 2000-2003 Tilo Christ. All Rights Reserved.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a 
 * copy of this software and associated documentation files (the  "Software"), 
 * to deal in the Software without restriction, including without limitation 
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the Software 
 * is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all 
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
 * INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
 * PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE 
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, 
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 **********************************************************************************/

#include "Customization.h"
#include <BmpGlue.h>
#include "Device.h"
#include "gfx/BlitterSupport.h"
#include "gfx/Direct4bppBlitter.h"
#include "gfx/Canvas.h"

using namespace prvrazor;


/*
 * TODO:
 * - Convert all routines to ASM for CW and GCC.
 * - Use Feature memory for bitmaps????
 */


Direct4bppTransparentBitmapArray::Direct4bppTransparentBitmapArray(const Canvas *canvas, UInt16 numBitmaps, UInt16 bitmapID, UInt16 maskID)
	: BitmapArray(numBitmaps), bitmaps(new Direct4bppTransparentBitmap[numBitmaps])
{
	// Determine bitmap extents
	{
		const MemHandle resH = DmGet1Resource( bitmapRsc, bitmapID);
		ErrNonFatalDisplayIf( !resH, "Missing bitmap" );
		BitmapType *const resP = static_cast<BitmapType *>(MemHandleLock(resH));
	
		BmpGlueGetDimensions(resP, &width, &height, NULL);
	
		MemPtrUnlock(resP);
		DmReleaseResource( resH );
	}

	// Prepare all bitmaps
	RectangleType bitmapRect;
	bitmapRect.topLeft.x = bitmapRect.topLeft.y = 0;
	bitmapRect.extent.x = width; bitmapRect.extent.y = height;
	
	for(UInt16 i = 0; i < numBitmaps; i++)
	{
		UInt16 error = errNone;
		Direct4bppTransparentBitmap *const bitmap = const_cast<Direct4bppTransparentBitmap *const>(&bitmaps[i]);

    	// Paint sprite-mask to offscreen window
		{
			const MemHandle resH = DmGet1Resource( bitmapRsc, maskID + i);
			ErrNonFatalDisplayIf( !resH, "Missing mask" );
			BitmapType *const resP = static_cast<BitmapType *>(MemHandleLock(resH));
		
			bitmap->mask = WinCreateOffscreenWindow(width, height, screenFormat, &error);
			if (error != errNone)
			{
				Device::panic("Not enough memory for mask.");
			}				
		
			WinSetDrawWindow(bitmap->mask);
			if (Device::supports35)
			{
				WinSetForeColor(15);
				WinSetBackColor(0);
			}		
			WinEraseRectangle(&bitmapRect, 0);
			WinDrawBitmap (resP, 0,0);
			MemPtrUnlock(resP);
			DmReleaseResource( resH );				
		}
	
		// Paint sprite-bitmap to offscreen window
		{
			const MemHandle resH = DmGet1Resource( bitmapRsc, bitmapID + i);
			ErrNonFatalDisplayIf( !resH, "Missing bitmap" );
			BitmapType *const resP = static_cast<BitmapType *>(MemHandleLock(resH));
		
			bitmap->bitmap = WinCreateOffscreenWindow(width, height, screenFormat, &error);
			if (error != errNone)
			{
				Device::panic("Not enough memory for bitmap.");
			}						
			
			WinSetDrawWindow(bitmap->bitmap);
			WinDrawBitmap (resP, 0,0);
			WinCopyRectangle(bitmap->mask, bitmap->bitmap, &bitmapRect, 0, 0, winErase);
		
			MemPtrUnlock(resP);
			DmReleaseResource( resH );
		}
	
		WinSetDrawWindow(canvas->drawWindow);
		
		bitmap->bitmapBits = (UInt8 *)(BlitterSupport::getWindowBits(bitmap->bitmap));
		bitmap->maskBits = (UInt8 *)(BlitterSupport::getWindowBits(bitmap->mask));
	}
}


Direct4bppTransparentBitmapArray::~Direct4bppTransparentBitmapArray()
{
	for (UInt16 i = 0; i < numBitmaps; i++)
	{
		const Direct4bppTransparentBitmap *const bitmap = &bitmaps[i];

		WinDeleteWindow(bitmap->bitmap, false);			
		WinDeleteWindow(bitmap->mask, false);	
	}

	delete [] bitmaps;
}



void Direct4bppTransparentBitmapArray::draw(const Canvas *canvas, UInt16 index, Coord x, Coord y) const
{
	const Direct4bppTransparentBitmap *const bitmap = &bitmaps[index];

	if (x & 1)
	{
		// Odd x-coordinate: Let PalmOS do the work...
		RectangleType size;
		size.topLeft.x = 0;
		size.topLeft.y = 0;
		size.extent.x = width;
		size.extent.y = height;			
	
		WinHandle drawWindow = WinGetDrawWindow();
		WinCopyRectangle(bitmap->mask, drawWindow, &size, x, y, winMask);
		WinCopyRectangle(bitmap->bitmap, drawWindow, &size, x, y, winOverlay);		
		
		return;
	}


// This is only C code, but already faster than the OS routines...


	// Determine memory address of bits
	UInt8 *foregroundBits = bitmap->bitmapBits;
	UInt8 *backgroundBits = bitmap->maskBits;

	UInt8 *drawBits = (UInt8 *)canvas->drawBits;

	// Clipping
	Int16 xOffset = 0;
	Int16 xExtent = width;
	if (x < 0)
	{
		xOffset = -x;
		xExtent -= xOffset;
	}
	
	if (x + width > canvas->width)
	{
		xExtent -= ((x + width) - canvas->width);
	}

	if (xExtent <= 0)
	{
		return;
	}
	

	Int16 yOffset = 0;
	Int16 yExtent = height;
	if (y < 0)
	{
		yOffset = 0 - y;
		yExtent -= yOffset;
	}
	
	if (y + height > canvas->height)
	{
		yExtent -= ((y + height) - canvas->height);
	}

	if (yExtent <= 0)
	{
		return;
	}
	

	// Paint
	{
		const UInt16 sourceOffset = (width * yOffset + xOffset) >> 1;
		foregroundBits += sourceOffset;
		backgroundBits += sourceOffset;	
	}
	
	drawBits += (canvas->width * (y + yOffset) + x + xOffset) >> 1;

	for (Int16 row = 0; row < yExtent; row++)
	{
		for (Int16 col = 0; col < (xExtent >> 1); col++)
		{
			UInt8 pixelByte = (*drawBits);
			pixelByte &= ~(*backgroundBits);
			pixelByte |= (*foregroundBits);
			(*drawBits) = pixelByte;
	
			foregroundBits++;
			backgroundBits++;
			drawBits++;
		}
	
		foregroundBits += (width - xExtent) >> 1;
		backgroundBits += (width - xExtent) >> 1;
		drawBits += (canvas->width - xExtent) >> 1;
	}
}

//---------------------------------------------------------------------------------------------------------


Direct4bppSolidBitmapArray::Direct4bppSolidBitmapArray(const Canvas */*canvas*/, UInt16 numBitmaps, UInt16 bitmapID)
	: BitmapArray(numBitmaps), bitmaps(new Direct4bppSolidBitmap[numBitmaps])
{
	for (UInt16 i = 0; i < numBitmaps; i++)
	{
		Direct4bppSolidBitmap *const bitmap = const_cast<Direct4bppSolidBitmap *const>(&bitmaps[i]);
	
		bitmap->bitmapResH = DmGet1Resource( bitmapRsc, bitmapID + i);
		bitmap->bitmap = static_cast<BitmapType *>(MemHandleLock(bitmap->bitmapResH));
		ErrNonFatalDisplayIf( !(bitmap->bitmap), "Missing bitmap" );		
	}

	// Determine size
	BmpGlueGetDimensions(bitmaps->bitmap, &width, &height, NULL);
}


Direct4bppSolidBitmapArray::~Direct4bppSolidBitmapArray()
{
	for (UInt16 i = 0; i < numBitmaps; i++)
	{
		const Direct4bppSolidBitmap *const bitmap = &bitmaps[i];
		
		MemPtrUnlock(bitmap->bitmap);
		DmReleaseResource(bitmap->bitmapResH);
	}

	delete [] bitmaps;
}



void Direct4bppSolidBitmapArray::draw(const Canvas */*canvas*/, UInt16 index, Coord x, Coord y) const
{
	const Direct4bppSolidBitmap *const bitmap = &bitmaps[index];

	// TODO: Should the canvas draw window be set?
	WinDrawBitmap(bitmap->bitmap, x, y);
}

//---------------------------------------------------------------------------------------------------------


Direct4bppSaveBufferArray::Direct4bppSaveBufferArray(UInt16 numBuffers, Coord width, Coord height)
    : SaveBufferArray(numBuffers, width, height), buffers(new Direct4bppSaveBuffer[numBuffers])
{
	// Round width to next multiple of 4
	UInt16 bufferWidth = width;
	if ((bufferWidth & 3) != 0)
	{
		bufferWidth += 4 - (bufferWidth & 3);
	}

	// Add another 4 bytes
	bufferWidth += 4;
	
	for (UInt16 i = 0; i < numBuffers; i++)
	{
		Direct4bppSaveBuffer *const buffer = &buffers[i];
		const UInt32 bufferSize = (bufferWidth * height) >> 1;
		buffer->bufferBits = static_cast<UInt8 *>(MemPtrNew(bufferSize));
		if (buffer->bufferBits == NULL)
		{
			Device::panic("Not enough memory for save buffer.");
		}		
	}
}


Direct4bppSaveBufferArray::~Direct4bppSaveBufferArray()
{
	for (UInt16 i = 0; i < numBuffers; i++)
	{
		const Direct4bppSaveBuffer *const buffer = &buffers[i];
		if (buffer->bufferBits != NULL)
		{
			MemPtrFree(buffer->bufferBits);
		}
	}
	
	delete [] buffers;
}


void Direct4bppSaveBufferArray::save(const Canvas *canvas, UInt16 index, Coord x, Coord y) const
{
	Direct4bppSaveBuffer *const buffer = &buffers[index];

	UInt8 *bufferBits = buffer->bufferBits;
	const UInt8 *drawBits = static_cast<UInt8 *>(canvas->drawBits);

	// Clipping
	Int16 xOffset = 0;
	Int16 xExtent = width;
	if (x < 0)
	{
		xOffset = -x;
		xExtent -= xOffset;
	}
	
	if (x + width > canvas->width)
	{
		xExtent -= ((x + width) - canvas->width);
	}

	if (xExtent <= 0)
	{
		buffer->colQuads = 0;
		return;
	}

	Int16 spanLeftX = x + xOffset;

	// Force even start address
	spanLeftX &= ~3;

	// Force xExtent divisible by 4
	if ((xExtent & 3) != 0)
	{
		xExtent += 4 - (xExtent & 3);
	}

	// Make sure, entire width is covered
	if (spanLeftX + xExtent < x + width)
	{
		xExtent &= ~3;
		xExtent += 4;
	}

	// Make sure covered span doesn't go off the right edge of Canvas
	if (spanLeftX + xExtent > canvas->width)
	{
		spanLeftX = canvas->width - xExtent;		
	}
	

	Int16 yOffset = 0;
	Int16 yExtent = height;
	if (y < 0)
	{
		yOffset = 0 - y;
		yExtent -= yOffset;
	}
	
	if (y + height > canvas->height)
	{
		yExtent -= ((y + height) - canvas->height);
	}

	if (yExtent <= 0)
	{
		buffer->yExtent = 0;
		return;
	}
	
	// Calculate low-level parameters
	const UInt32 drawOffset = (canvas->width * (y + yOffset) + spanLeftX) >> 1;
	UInt16 colQuads = xExtent >> 2;
	const UInt32 canvasRemainder = (canvas->width - xExtent) >> 1;

	// Save low-level parameters in buffer metadata
	buffer->canvasRemainder = canvasRemainder;
	buffer->drawOffset = drawOffset;
	buffer->colQuads = colQuads;
	buffer->yExtent = yExtent;

	// Save drawbits to buffer
	drawBits += drawOffset;

#ifdef MODE_M68K

#ifdef __GNUC__
	yExtent--;
	colQuads--;

	asm("
saverowloop:
	move.w %3, %%d2;

savecolloop:
	move.w (%1)+, (%0)+;
dbra.w  %%d2, savecolloop;

	add.l   %4, %1;
	
dbra.w  %2, saverowloop;
	" : 
	  : "a" (bufferBits), "a" (drawBits), "d" (yExtent), "r" (colQuads), "r" (canvasRemainder)
	  : "%d2");
#else
	// Codewarrior
	saveM68KAsm(bufferBits, drawBits, yExtent, colQuads, canvasRemainder);
#endif
#else
	// Unknown processor type
	for (UInt16 row = 0; row < yExtent; row++)
	{
		for (UInt16 col = 0; col < colQuads; col++)
		{
			(*((UInt16 *)bufferBits)) = (*((UInt16 *)drawBits));		
			bufferBits += 2;
			drawBits += 2;
		}

		drawBits += canvasRemainder;
	}
#endif
}


void Direct4bppSaveBufferArray::restore(const Canvas *canvas, UInt16 index) const
{
	const Direct4bppSaveBuffer *const buffer = &buffers[index];

	const UInt8 *bufferBits = buffer->bufferBits;
	UInt8 *drawBits = static_cast<UInt8 *>(canvas->drawBits);

	// Restore low-level metadata from buffer
	const UInt32 drawOffset = buffer->drawOffset;
	UInt16 colQuads = buffer->colQuads;
	const UInt32 canvasRemainder = buffer->canvasRemainder;
	Int16 yExtent = buffer->yExtent;

	if ((colQuads == 0) || (yExtent == 0))
	{
		return;
	}

	// Restore buffer to drawing canvas
	drawBits += drawOffset;

#ifdef MODE_M68K

#ifdef __GNUC__
	yExtent--;
	colQuads--;

	asm("
restorerowloop:
	move.w %3, %%d2;

restorecolloop:
	move.w (%0)+, (%1)+;
dbra.w  %%d2, restorecolloop;

	add.l   %4, %1;
	
dbra.w  %2, restorerowloop;
	" : 
	  : "a" (bufferBits), "a" (drawBits), "d" (yExtent), "r" (colQuads), "r" (canvasRemainder)
	  : "%d2");
#else
	// Codewarrior
	restoreM68KAsm(bufferBits, drawBits, yExtent, colQuads, canvasRemainder);
#endif
#else
	// Unknown processor
	for (UInt16 row = 0; row < yExtent; row++)
	{
		for (UInt16 col = 0; col < colQuads; col++)
		{
			(*((UInt16 *)drawBits)) = (*((UInt16 *)bufferBits));		
			bufferBits += 2;
			drawBits += 2;
		}

		drawBits += canvasRemainder;
	}
#endif
}



#ifdef __MWERKS__
asm void Direct4bppSaveBufferArray::saveM68KAsm(
	void *bufferBits, const void *drawBits,
	Int16 yExtent, UInt16 colQuads, UInt32 canvasRemainder
	) const
{
	fralloc +

	movem.l D0-D3/A0-A1, -(SP)

	move.l bufferBits, A0
	move.l drawBits, A1

	move.w yExtent, D0
	subq.w #1, D0
	move.w colQuads, D1
	subq.w #1, D1
	move.l canvasRemainder, D3	

rowloop:
	move.w D1, D2

colloop:
	move.w (A1)+, (A0)+

dbra.w  D2, colloop

	add.l 	D3, A1
	
dbra.w  D0, rowloop

	movem.l (SP)+, D0-D3/A0-A1

	frfree

	rts	
}


asm void Direct4bppSaveBufferArray::restoreM68KAsm(
	const void *bufferBits, void *drawBits,
	Int16 yExtent, UInt16 colQuads, UInt32 canvasRemainder
	) const
{
	fralloc +

	movem.l D0-D3/A0-A1, -(SP)

	move.l bufferBits, A0
	move.l drawBits, A1

	move.w yExtent, D0
	subq.w #1, D0
	move.w colQuads, D1
	subq.w #1, D1
	move.l canvasRemainder, D3	

rowloop:
	move.w D1, D2

colloop:
	move.w (A0)+, (A1)+

dbra.w  D2, colloop

	add.l 	D3, A1
	
dbra.w  D0, rowloop

	movem.l (SP)+, D0-D3/A0-A1

	frfree

	rts	
}
#endif


//---------------------------------------------------------------------------------------------------------


Direct4bppBlitter::Direct4bppBlitter(Canvas *canvas) : Blitter(canvas)
{
}


Direct4bppBlitter::~Direct4bppBlitter()
{
}


const BitmapArray *Direct4bppBlitter::prepareTransparentBitmaps(UInt16 numBitmaps, DmResID bitmapID, DmResID maskID) const
{
	return new Direct4bppTransparentBitmapArray(canvas, numBitmaps, bitmapID, maskID);
}


const BitmapArray *Direct4bppBlitter::prepareSolidBitmaps(UInt16 numBitmaps, DmResID bitmapID) const
{
	return new Direct4bppSolidBitmapArray(canvas, numBitmaps, bitmapID);
}



void Direct4bppBlitter::drawText(Color textColor, Color bgColor, FontID font, const Char *chars, Int16 len, Coord x, Coord y) const
{
	canvas->setForegroundColor(textColor);
	canvas->setBackgroundColor(bgColor);

	FntSetFont(font);	
	WinDrawChars(chars, len, x, y);
}	



void Direct4bppBlitter::fillRectangle(Color color, Coord topX, Coord topY, Coord bottomX, Coord bottomY) const
{
	canvas->setForegroundColor(color);
	
	RectangleType rectangle;
	rectangle.topLeft.x = topX;
	rectangle.topLeft.y = topY;
	rectangle.extent.x = (bottomX - topX) + 1;
	rectangle.extent.y = (bottomY - topY) + 1;
	
	WinDrawRectangle(&rectangle, 0);
}


SaveBufferArray *Direct4bppBlitter::prepareSaveBuffers(UInt16 numBuffers, Coord width, Coord height) const
{
    return new Direct4bppSaveBufferArray(numBuffers, width, height);
}
