/*********************************************************************************
 *
 * Razor! Engine - A modular C++ presentation engine
 *
 * $Id: SonyClieBlitter.cpp,v 1.4 2003/05/29 15:45:25 teacy Exp $
 *
 * Copyright (c) 2000-2003 Tilo Christ. All Rights Reserved.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a 
 * copy of this software and associated documentation files (the  "Software"), 
 * to deal in the Software without restriction, including without limitation 
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the Software 
 * is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all 
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
 * INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
 * PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE 
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, 
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 **********************************************************************************/

/*
 * SEE:
 * - http://www.palmos.com/dev/support/docs/recipes/copying_bitmap.html
 * - http://www.palmos.com/dev/support/docs/recipes/using_winpaintbitmap.html
 * 
 * On the Clie, I originally assumed that the OS is faster than assembly code, because
 * it makes use of the MediaQ 1100 hardware blitter. But a user reported that the 
 * Direct8bppBlitter was a great deal faster in 8bpp lo-res mode. SonyClieCanvas makes
 * more appropriate blitter choices now.
 */


#include "Customization.h"

#ifdef SUPPORT_CLIE
// Duplicate definition: One in OS5 SDK, one in Sony SDK
#undef vchrSonyMin
#undef vchrSonyMax
#include <BmpGlue.h>
#include <SonyCLIE.h>
#include "gfx/SonyClieBlitter.h"
#include "gfx/SonyClieCanvas.h"
#include "gfx/BlitterSupport.h"
#include "gfx/Display.h"
#include "gfx/Canvas.h"
#include "Device.h"

using namespace prvrazor;



SonyClieTransparentBitmapArray::SonyClieTransparentBitmapArray(const Canvas *canvas, UInt16 numBitmaps, UInt16 bitmapID, UInt16 maskID)
	: BitmapArray(numBitmaps), bitmaps(new SonyClieTransparentBitmap[numBitmaps])
{
	const SonyClieCanvas *const sonyCanvas = static_cast<const SonyClieCanvas *>(canvas);
	const UInt16 hrRefnum = sonyCanvas->hrRefnum;

	UInt16 error = errNone;

    Boolean v3Bitmaps = false;
	// Find bitmap and determine bitmap extents
	if (Display::density == Display::DENSITY_DOUBLE)
	{
        // Need to convert V3 bitmap to V2 bitmap
		MemHandle resH = DmGet1Resource( bitmapRsc, bitmapID);
		ErrNonFatalDisplayIf( !resH, "Missing transparent bitmap" );
		BitmapType *resP = static_cast<BitmapType *>(MemHandleLock(resH));
		const BitmapType* highDensityP = BlitterSupport::getHighDensityMember(resP, Display::depth);
		ErrNonFatalDisplayIf(highDensityP == NULL, "No high density member" );						
		BmpGlueGetDimensions(highDensityP, &width, &height, NULL);
		MemPtrUnlock(resP);
		DmReleaseResource( resH );
        v3Bitmaps = true;
	}
	else
	{
		// Get lo-res bitmap
		MemHandle resH = DmGet1Resource( bitmapRsc, bitmapID);
		ErrNonFatalDisplayIf( !resH, "Missing transparent bitmap" );
		BitmapType *resP = static_cast<BitmapType *>(MemHandleLock(resH));
		BmpGlueGetDimensions(resP, &width, &height, NULL);
		MemPtrUnlock(resP);
		DmReleaseResource( resH );
	}
	
	RectangleType bitmapRect;
	bitmapRect.topLeft.x = bitmapRect.topLeft.y = 0;
	bitmapRect.extent.x = width; bitmapRect.extent.y = height;

	// Prepare all bitmaps
	for (UInt16 i = 0; i < numBitmaps; i++)
	{
		SonyClieTransparentBitmap *const bitmap = const_cast<SonyClieTransparentBitmap *const>(&bitmaps[i]);
	
		// Paint sprite-mask to offscreen window	
		bitmap->mask = HRWinCreateOffscreenWindow(hrRefnum, width, height, screenFormat, &error);
	
		WinSetDrawWindow(bitmap->mask);
		WinSetForeColor((1 << Display::depth) - 1);
		WinSetBackColor(0);	
		HRWinEraseRectangle(hrRefnum, &bitmapRect, 0);

		if (v3Bitmaps)
		{
			const MemHandle resH = DmGet1Resource( bitmapRsc, maskID + i);
			ErrNonFatalDisplayIf( !resH, "Missing mask" );
			BitmapType *const resP = static_cast<BitmapType *>(MemHandleLock(resH));
			BitmapType *const hiresP = BlitterSupport::createIndirectV2Copy(resP, 1);
			HRWinDrawBitmap (hrRefnum, hiresP, 0,0);
			MemPtrFree(hiresP);
			MemPtrUnlock(resP);
			DmReleaseResource( resH );				
		}
		else
		{
			const MemHandle resH = DmGet1Resource( bitmapRsc, maskID + i);
			ErrNonFatalDisplayIf( !resH, "Missing mask" );
			BitmapType *const resP = static_cast<BitmapType *>(MemHandleLock(resH));
			HRWinDrawBitmap (hrRefnum, resP, 0,0);
			MemPtrUnlock(resP);
			DmReleaseResource( resH );				
		}


		// Paint sprite-bitmap to offscreen window
		bitmap->bitmap = HRWinCreateOffscreenWindow(hrRefnum, width, height, screenFormat, &error);
		WinSetDrawWindow(bitmap->bitmap);
		if (v3Bitmaps)
		{
			const MemHandle resH = DmGet1Resource( bitmapRsc, bitmapID + i);
			ErrNonFatalDisplayIf( !resH, "Missing bitmap" );
			BitmapType *const resP = static_cast<BitmapType *>(MemHandleLock(resH));
			BitmapType *const hiresP = BlitterSupport::createIndirectV2Copy(resP, Display::depth);
			HRWinDrawBitmap (hrRefnum, hiresP, 0,0);
			MemPtrFree(hiresP);
			MemPtrUnlock(resP);
			DmReleaseResource( resH );				
		}
		else
		{
			const MemHandle resH = DmGet1Resource( bitmapRsc, bitmapID + i);
			ErrNonFatalDisplayIf( !resH, "Missing bitmap" );
			BitmapType *const resP = static_cast<BitmapType *>(MemHandleLock(resH));
			HRWinDrawBitmap (hrRefnum, resP, 0,0);
			MemPtrUnlock(resP);
			DmReleaseResource( resH );				
		}

		HRWinCopyRectangle(hrRefnum, bitmap->mask, bitmap->bitmap, &bitmapRect, 0, 0, winErase);
	
		WinSetDrawWindow(canvas->drawWindow);
	}
}


SonyClieTransparentBitmapArray::~SonyClieTransparentBitmapArray()
{
	for (UInt16 i = 0; i < numBitmaps; i++)
	{
		const SonyClieTransparentBitmap *const bitmap = &bitmaps[i];
		WinDeleteWindow(bitmap->bitmap, false);			
		WinDeleteWindow(bitmap->mask, false);	
	}

	delete [] bitmaps;
}


void SonyClieTransparentBitmapArray::draw(const Canvas *canvas, UInt16 index, Coord x, Coord y) const
{
	const SonyClieTransparentBitmap *const bitmap = &bitmaps[index];

	const Int8 shiftFactor = canvas->shiftFactor;

	RectangleType size;
	size.topLeft.x = 0;
	size.topLeft.y = 0;
	size.extent.x = width << shiftFactor;
	size.extent.y = height << shiftFactor;			

	x <<= shiftFactor;
	y <<= shiftFactor;

	const UInt16 hrRefnum = static_cast<const SonyClieCanvas *>(canvas)->hrRefnum;
	const WinHandle drawWindow = canvas->drawWindow;
	HRWinCopyRectangle(hrRefnum, bitmap->mask, drawWindow, &size, x, y, winMask);
	HRWinCopyRectangle(hrRefnum, bitmap->bitmap, drawWindow, &size, x, y, winOverlay);		
}



SonyClieSolidBitmapArray::SonyClieSolidBitmapArray(const Canvas */*canvas*/, UInt16 numBitmaps, UInt16 bitmapID)
	: BitmapArray(numBitmaps), bitmaps(new SonyClieSolidBitmap[numBitmaps])
{
	Boolean v3Bitmaps = false;
	
	if (Display::density == Display::DENSITY_DOUBLE)
	{
		// Need to convert V3 bitmap to V2 bitmap
		const MemHandle resH = DmGet1Resource( bitmapRsc, bitmapID);
		ErrNonFatalDisplayIf( !resH, "Missing solid bitmap" );
		BitmapType *const resP = static_cast<BitmapType *>(MemHandleLock(resH));
		const BitmapType *const highDensityP = BlitterSupport::getHighDensityMember(resP, Display::depth);
		ErrNonFatalDisplayIf(highDensityP == NULL, "No high density member" );						
		BmpGlueGetDimensions(highDensityP, &width, &height, NULL);
		MemPtrUnlock(resP);
		DmReleaseResource( resH );
		v3Bitmaps = true;
	}
	else
	{
		// Get lo-res bitmap
		const MemHandle resH = DmGet1Resource( bitmapRsc, bitmapID);
		ErrNonFatalDisplayIf( !resH, "Missing solid bitmap" );
		BitmapType *const resP = static_cast<BitmapType *>(MemHandleLock(resH));
		BmpGlueGetDimensions(resP, &width, &height, NULL);
		MemPtrUnlock(resP);
		DmReleaseResource( resH );
	}


	if (v3Bitmaps)
	{
		for (UInt16 i = 0; i < numBitmaps; i++)
		{	
			SonyClieSolidBitmap *const bitmap = const_cast<SonyClieSolidBitmap *const>(&bitmaps[i]);
	
			bitmap->bitmapResH = DmGet1Resource( bitmapRsc, bitmapID + i);
			bitmap->bitmap = static_cast<BitmapType *>(MemHandleLock(bitmap->bitmapResH));
			bitmap->v3Bitmap = BlitterSupport::createIndirectV2Copy(bitmap->bitmap, Display::depth);
			ErrNonFatalDisplayIf( !(bitmap->bitmap), "Missing solid bitmap" );		
		}
	}
	else
	{
		for (UInt16 i = 0; i < numBitmaps; i++)
		{	
			SonyClieSolidBitmap *const bitmap = const_cast<SonyClieSolidBitmap *const>(&bitmaps[i]);
	
			bitmap->bitmapResH = DmGet1Resource( bitmapRsc, bitmapID + i);
			bitmap->bitmap = static_cast<BitmapType *>(MemHandleLock(bitmap->bitmapResH));
			bitmap->v3Bitmap = NULL;
			ErrNonFatalDisplayIf( !(bitmap->bitmap), "Missing solid bitmap" );		
		}
	}
}



SonyClieSolidBitmapArray::~SonyClieSolidBitmapArray()
{
	for (UInt16 i = 0; i < numBitmaps; i++)
	{
		const SonyClieSolidBitmap *const bitmap = &bitmaps[i];
		MemPtrUnlock(bitmap->bitmap);
		if (bitmap->v3Bitmap)
		{
			MemPtrUnlock(bitmap->v3Bitmap);
		}
		DmReleaseResource(bitmap->bitmapResH);
	}

	delete [] bitmaps;
}


void SonyClieSolidBitmapArray::draw(const Canvas *canvas, UInt16 index, Coord x, Coord y) const
{
	const Int8 shiftFactor = canvas->shiftFactor;
	const UInt16 hrRefnum = static_cast<const SonyClieCanvas *>(canvas)->hrRefnum;

	const SonyClieSolidBitmap *const bitmap = &bitmaps[index];

	if (bitmap->v3Bitmap != NULL)
	{
		HRWinDrawBitmap(hrRefnum, bitmap->v3Bitmap, x << shiftFactor, y << shiftFactor);
	}
	else
	{
		HRWinDrawBitmap(hrRefnum, bitmap->bitmap, x << shiftFactor, y << shiftFactor);
	}	
}




SonyClieSaveBufferArray::SonyClieSaveBufferArray(const Canvas *canvas, UInt16 numBuffers, Coord width, Coord height)
    : SaveBufferArray(numBuffers, width, height), buffers(new SonyClieSaveBuffer[numBuffers])
{
    const Int8 shiftFactor = canvas->shiftFactor;

	this->width = width << shiftFactor;
	this->height = height << shiftFactor;

	const SonyClieCanvas *const sonyCanvas = static_cast<const SonyClieCanvas *>(canvas);
	for (UInt16 i = 0; i < numBuffers; i++)
	{
		SonyClieSaveBuffer *const buffer = &buffers[i];
		Err error;
		buffer->buffer = HRWinCreateOffscreenWindow(sonyCanvas->hrRefnum, width, height, screenFormat, &error);
		if (error != errNone)
		{
			Device::panic("Not enough memory for save buffer.");
		}		
	}
}


SonyClieSaveBufferArray::~SonyClieSaveBufferArray()
{
	for (UInt16 i = 0; i < numBuffers; i++)
	{
		const SonyClieSaveBuffer *const buffer = &buffers[i];		
		WinDeleteWindow(buffer->buffer, false);			
	}
	
	delete [] buffers;
}


void SonyClieSaveBufferArray::save(const Canvas *canvas, UInt16 index, Coord x, Coord y) const
{
    const Int8 shiftFactor = canvas->shiftFactor;
    
	x <<= shiftFactor;
	y <<= shiftFactor;

	RectangleType saveBounds;
	saveBounds.topLeft.x = x;
	saveBounds.topLeft.y = y;
	// Extent size already adjusted by constructor
	saveBounds.extent.x = width;
	saveBounds.extent.y = height;

	SonyClieSaveBuffer *const buffer = &buffers[index];		
	buffer->x = x;
	buffer->y = y;

	HRWinCopyRectangle(static_cast<const SonyClieCanvas *>(canvas)->hrRefnum, canvas->drawWindow, buffer->buffer, &saveBounds, 0, 0, winPaint);
}


void SonyClieSaveBufferArray::restore(const Canvas *canvas, UInt16 index) const
{
	const SonyClieCanvas *const sonyCanvas = static_cast<const SonyClieCanvas *>(canvas);

	// All coordinates already adjusted

	RectangleType srcRect;
	srcRect.topLeft.x = 0;
	srcRect.topLeft.y = 0;
	srcRect.extent.x = width;
	srcRect.extent.y = height;
	
	const SonyClieSaveBuffer *const buffer = &buffers[index];		

	HRWinCopyRectangle(sonyCanvas->hrRefnum, buffer->buffer, canvas->drawWindow, &srcRect, buffer->x, buffer->y, winPaint);
}




SonyClieBlitter::SonyClieBlitter(const SonyClieCanvas *canvas) 
	: Blitter(canvas)
{
	this->hrRefnum = canvas->hrRefnum;
}


SonyClieBlitter::~SonyClieBlitter()
{
}


const BitmapArray *SonyClieBlitter::prepareTransparentBitmaps(UInt16 numBitmaps, DmResID bitmapID, DmResID maskID) const
{
	return new SonyClieTransparentBitmapArray(canvas, numBitmaps, bitmapID, maskID);
}


const BitmapArray *SonyClieBlitter::prepareSolidBitmaps(UInt16 numBitmaps, DmResID bitmapID) const
{
	return new SonyClieSolidBitmapArray(canvas, numBitmaps, bitmapID);
}


void SonyClieBlitter::drawText(Color textColor, Color bgColor, FontID font, const Char *chars, Int16 len, Coord x, Coord y) const
{
	canvas->setBackgroundColor(bgColor);
	canvas->setForegroundColor(textColor);

    const Int8 shiftFactor = canvas->shiftFactor;
	HRFntSetFont(hrRefnum, (HRFontID)(((shiftFactor == 1) ? 8 : 0) + font));
	HRWinDrawChars(hrRefnum, chars, len, x << shiftFactor, y << shiftFactor);
}	


void SonyClieBlitter::fillRectangle(Color color, Coord topX, Coord topY, Coord bottomX, Coord bottomY) const
{
	canvas->setForegroundColor(color);
	
    const Int8 shiftFactor = canvas->shiftFactor;

	RectangleType rectangle;
	rectangle.topLeft.x = topX << shiftFactor;
	rectangle.topLeft.y = topY << shiftFactor;
	rectangle.extent.x = ((bottomX - topX) + 1) << shiftFactor;
	rectangle.extent.y = ((bottomY - topY) + 1) << shiftFactor;
	
	HRWinDrawRectangle(hrRefnum, &rectangle, 0);
}


SaveBufferArray *SonyClieBlitter::prepareSaveBuffers(UInt16 numBuffers, Coord width, Coord height) const
{
    return new SonyClieSaveBufferArray(canvas, numBuffers, width, height);
}
#endif // SUPPORT_CLIE
