/*********************************************************************************
 *
 * Razor! Engine - A modular C++ presentation engine
 *
 * $Id: Presentation.h,v 1.3 2003/05/31 20:05:40 teacy Exp $
 *
 * Copyright (c) 2000-2003 Tilo Christ. All Rights Reserved.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a 
 * copy of this software and associated documentation files (the  "Software"), 
 * to deal in the Software without restriction, including without limitation 
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the Software 
 * is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all 
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
 * INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
 * PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE 
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, 
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 **********************************************************************************/

/*
 * TODO: -Figure out why moving Presentation to namespace prvrazor won't compile.
 */

#ifndef PRESENTATION_H
#define PRESENTATION_H

#include <PalmOS.h>
#include "RazorSections.h"
#include "Customization.h"


class Display;
class ActionEngine;
class Screen;


/**
 * Presentation controls the flow of a presentation. It is a facade
 * that delegates most of the work to other classes.
 *
 * @see ActionEngine
 * @see Canvas
 * @see HardKeyManager
 */
class Presentation
{
	public:

///@name Construction / Destruction
//@{
	/**
	 * Construct a new Presentation.
	 */
	Presentation() SEC_RAZOR_INIT;
	
	/**
	 * Destroy the Presentation.
	 */
	~Presentation() SEC_RAZOR_INIT;

//@}

	//********************************************************************************************************
	
	
/// @name Flow Control & Timing
//@{

	/**
	 * Begin the presentation.
	 */
	void begin() SEC_RAZOR_INIT;

	/**
	 * Has the presentation begun?
	 */
	Boolean hasBegun() const;
	
	/**
	 * Step to the next period.
	 */
	void nextPeriod();


	/**
	 * Draw the state of the current period again.
	 */
	void redraw();

	
	/**
	 * Is the presentation currenty paused?
	 *
	 * @return true if paused, false if not paused
	 */
	Boolean isPaused() const;


	/**
	 * Pause the presentation.
	 */
	void pause() SEC_RAZOR_INIT;


	/**
	 * Resume a paused presentation.
	 */
	void resume() SEC_RAZOR_INIT;
	

	/**
	 * Determine the amount of time until the next period begins.
	 * The result can be passed directly into a call to EvtGetEvent().
	 *
	 * @return the amount of time in ticks (1/100 secs) or evtWaitForever
	 */
	Int32 getTimeUntilNextPeriod() const;
//@}


/// @name Input management
//@{
	/**
	 * Receive an input event. The event will be forwarded to
	 * the ActionEngine which has to decide whether it wants to
	 * react to it.
	 */
	void receiveEvent(const EventType* evtPtr);
	
//@}

    /**
     * Indicates whether the presentation currently wants a menu.
     *
     * @return true = menu enabled, false = menu disabled
     */
    bool enableMenu();


	//********************************************************************************************************
	private:


/// @name Multimedia output
//@{
	/**
	 * Draw the current state of the presentation. Drawing will occur into
	 * an invisible buffer.
	 */
	void draw();
	
	
	/**
	 * Show the result of the last invocation of draw(). 
	 */
	void show();
//@}


	/**
	 * Switch over to the next screen.
	 */
	void switchScreen();


	//********************************************************************************************************
		
	/// The display
	Display& display;

	/// The ActionEngine
	ActionEngine& actionEngine;

	/// The current screen
	Screen *currentScreen;

	/// Current state of the presentation
	enum 
	{
		presReady,              ///<The presentation is ready, but has not begun yet.
		presResuming, 			///<Don't change the current state. Wait a little before setting presentation in motion.
		presRunning, 	 		///<Advance and draw the current state
		presPaused              ///<Presentation paused. Don't change the current state.
	} state;


	UInt32					nextPeriodTime;		/// Time when next period occurs
	UInt32					pausedTime;			/// When did the pause start?

	UInt32					advanceTimeInterval;
	UInt32					pauseLengthBeforeResuming;

	/// Current drawing state of the presentation
	enum
	{
		imgCorrupt,             ///<The current state of the presentation has not been drawn, and is not on the screen
		imgViewportReady,		///<The current state of the presentation has been drawn into the viewport
		imgViewportShowing		///<The current state of the presentation is on the screen
	} imgState;
};


#endif
