/*********************************************************************************
 *
 * Razor! Engine - A modular C++ presentation engine
 *
 * $Id: SoundEngine.h,v 1.1 2003/01/26 10:46:47 teacy Exp $
 *
 * Copyright (c) 2000-2003 Tilo Christ. All Rights Reserved.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a 
 * copy of this software and associated documentation files (the  "Software"), 
 * to deal in the Software without restriction, including without limitation 
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the Software 
 * is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all 
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
 * INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
 * PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE 
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, 
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 **********************************************************************************/



#ifndef SOUND_ENGINE_H
#define SOUND_ENGINE_H

#include <PalmOS.h>
#include "Sections.h"
#include "Customization.h"


namespace prvrazor
{
class MusicEngine;
class FxEngine;
}



/**
 * SoundEngine is a facade for sound output from two different sources (music, FX).
 */
class SoundEngine
{
	public:		
	/**
	 * Play a sound effect which is stored in a resource of type 'Tsfx' with the
	 * specified resource ID.
     *
     * @param resID the ID of the 'Tsfx' resource
	 */
	static void playFx(DmResID resID);
	
	/**
	 * Play a song where the sequence of patterns is stored in a resource of type
	 * 'Ttrk' with the specified resource ID. 
	 * The patterns themselves are stored in resources of type 'Tpat' where the 
	 * pattern ID is the resource ID.
     *
     * @param resID the ID of the 'Ttrk' resource
	 */
	static void playSong(DmResID resID);
	
	
	/**
	 * Stop the currently playing song.
	 */
	static void stopSong();
	

	/**
	 * This method needs to be invoked periodically, in order for SoundEngine
	 * to work.
	 */
	static void timeTick();


	private:	
	// SoundEngine cannot be instantiated or destroyed.
	SoundEngine() {}	
	~SoundEngine() {}

	static void init() SEC_RAZOR_INIT;
	static void destroy() SEC_RAZOR_INIT;

	static prvrazor::FxEngine *fxEngine;
	static prvrazor::MusicEngine *musicEngine;

	static UInt16 baseAmplitude;
	static Boolean mute;
	static UInt32 nextTimer;


	friend class Presentation;
};



namespace prvrazor
{

/**
 * FxEngine can play sound fx.
 */
class FxEngine
{
	private:
	
	FxEngine(UInt16 baseAmplitude) SEC_RAZOR_INIT;
	~FxEngine() SEC_RAZOR_INIT;
	
	/**
	 * Play the next portion of the sound.
     *
	 * @return true = sound has been played, false = no sound has been played
	 */
	Boolean timeTick();
	
	void playFx(DmResID resID);


	Boolean playing;
	DmResID fxResID;
	UInt16 fxPosition;
	UInt16 baseAmplitude;

	/// A frequency with value FREQ_END means, that the end of the sound effect has been reached.
	static const UInt16 FREQ_END = 0x0000;
	/// A duration with value DURATION_YIELD means, that timeTick() shall return control to the caller.
	static const UInt8 DURATION_YIELD = 0x00;

	friend class SoundEngine;
};


/**
 * MusicEngine can playback music. It is based on a pattern model,
 * comparable to that of MOD players, and it can fake three channels.
 */
class MusicEngine
{
	private:

	MusicEngine(UInt16 baseAmplitude) SEC_RAZOR_INIT;
	~MusicEngine() SEC_RAZOR_INIT;


	/**
	 * Play the next portion of the sound.
     *
	 * @return true = sound has been played, false = no sound has been played
	 */
	Boolean timeTick();

	void playSong(DmResID resID) SEC_RAZOR_INIT;
	void stopSong() SEC_RAZOR_INIT;

	void setPatternFromTrack();
	void releasePattern();

	void playNote(UInt16 note, UInt16 octave, UInt16 amplitude, UInt16 duration, Boolean wait) const;
	void playMute(UInt16 duration, Boolean wait) const;

	void hitIt(UInt16 channel, UInt16 amplitude, UInt16 duration, Boolean wait) const;


	DmResID trackResID;
	UInt16 trackPosition;
	
	UInt16 pattern;
	UInt16 patternPosition;
	Char* patternString;
	MemHandle patternResH;

	Boolean playing;
	Boolean channel1Playing;
	Boolean channel2Playing;
	Boolean channel3Playing;
	UInt16 playingChannel;

	UInt16 baseDuration;
	UInt16 baseAmplitude;

	static const UInt8 CHANNEL1_MUTE = 0x01;
	static const UInt8 CHANNEL2_MUTE = 0x02;
	static const UInt8 CHANNEL3_MUTE = 0x04;
	static const UInt8 TRACK_REPEAT = 0xFE;
	static const UInt8 TRACK_END    = 0xFF;

	friend class SoundEngine;
};
}

#endif
